# sage_setup: distribution = sagemath-graphs
from sage.misc.lazy_import import lazy_import

lazy_import('sage.sandpiles.sandpile', 'Sandpile')
lazy_import('sage.sandpiles.sandpile', 'SandpileDivisor')
lazy_import('sage.sandpiles.sandpile', 'SandpileConfig')
lazy_import('sage.sandpiles.sandpile', 'firing_graph')
lazy_import('sage.sandpiles.sandpile', 'parallel_firing_graph')
lazy_import('sage.sandpiles.sandpile', 'wilmes_algorithm')
lazy_import('sage.sandpiles.sandpile', 'triangle_sandpile')

lazy_import('sage.sandpiles.examples', 'sandpiles')
del lazy_import
